/**
 * This program has been developed by a student from the bachelor Information Science at
 * Utrecht University within the Research Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

import React, { useState, useEffect, useMemo } from 'react';
import { NodeLinkViewModel } from './NodeLinkViewModel';
import { useGraphQueryResult } from '../../data-access/store';
import { MapView, DeckGL } from 'deck.gl/typed';
import { getTooltip } from './Tooltip';
import { Node } from './types';
import * as d3 from 'd3';
import { generateBaseLayer } from './layers/BaseLayer';
import { FacetingLayer } from './layers/FacetingLayer';
import { ClusterLayer } from './layers/ClusterLayer';
import { NodeLinkLayer } from './layers/NodeLinkLayer';

type Props = {
  mapType: number;
  nodelink: string;
  isAggregated: boolean;
  isFanned: boolean;
  brushing: boolean;
};

export const NodeLinkMap = React.memo(
  ({
    mapType = 0,
    nodelink,
    isAggregated,
    isFanned,
    brushing = false,
  }: Props) => {
    const [graph, setGraph] = useState<NodeLinkViewModel | null>(null);
    const [isLoading, setIsLoading] = useState<boolean>(true);
    const [viewport, setViewport] = useState<{ [key: string]: number }>({
      longitude: -102.4,
      latitude: 37.8,
      zoom: 3,
    });

    const graphQueryResult = useGraphQueryResult();

    useEffect(() => {
      if (graphQueryResult) {
        const graphModel = new NodeLinkViewModel({
          nodes: graphQueryResult.nodes,
          edges: graphQueryResult.edges,
        });
        setGraph(graphModel);
        setIsLoading(false);
      }
    }, [graphQueryResult]);

    const onViewStateChange = (viewState) => {
      setViewport(viewState);
    };

    const edges = useMemo(() => graph?.getEdges(), [graph?.data]);
    const nodes = useMemo(() => graph?.getNodes(), [graph?.data]);

    const colorScale: d3.ScaleLinear<string, string> = d3
      .scaleLinear<string>()
      .domain([0, 6000])
      .range(['blue', 'red']);

    const layers = [generateBaseLayer(mapType)];

    if (nodelink === 'cluster') {
      layers.push(
        new ClusterLayer({
          graph: graph,
          edges: edges,
          nodes: nodes,
          colorScale: colorScale,
          brushing: brushing,
        })
      );
    } else if (nodelink === 'faceting') {
      layers.push(
        new FacetingLayer({
          graph: graph,
          edges: edges,
          nodes: nodes,
          colorScale: colorScale,
          isAggregated: isAggregated,
          isFanned: isFanned,
          brushing: brushing,
        })
      );
    } else {
      layers.push(
        new NodeLinkLayer({
          edges: edges,
          nodes: nodes,
          colorScale: colorScale,
          brushing: brushing,
        })
      );
    }

    return (
      <div>
        {!isLoading && (
          <DeckGL
            layers={layers}
            views={new MapView({ repeat: true, controller: true })}
            initialViewState={viewport}
            getTooltip={getTooltip}
            onViewStateChange={({ viewState }) => onViewStateChange(viewState)}
          />
        )}
      </div>
    );
  }
);

export default NodeLinkMap;
