/**
 * This program has been developed by a student from the bachelor Information Science at
 * Utrecht University within the Research Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

import { CompositeLayer } from 'deck.gl/typed';
import Supercluster from 'supercluster';
import { ScatterplotLayer, LineLayer } from '@deck.gl/layers/typed';
import { LayerProps } from '../types';
import { calcNodeDistance, getColor } from '../utils';
import { BrushingExtension } from '@deck.gl/extensions/typed';
import { EdgeCrossings } from '../readability';

export class ClusterLayer extends CompositeLayer<LayerProps> {
  /**
   * Layer that clusters the given graph
   * Cluster layer uses supercluster library to cluster nodes
   * Edges get updated accordingly
   */
  static componentName = 'Cluster Layer';

  shouldUpdateState({ changeFlags }) {
    return changeFlags.somethingChanged;
  }

  updateState({ props, oldProps, changeFlags }) {
    const rebuildIndex = changeFlags.dataChanged;
    const zoom = Math.floor(this.context.viewport.zoom);

    if (rebuildIndex) {
      const index = new Supercluster({
        radius: 40,
        maxZoom: 16,
        map: (props) => ({ ids: [props.id] }),
        reduce: (accumulated, props) => {
          accumulated.ids.push(...props.ids);
          return accumulated;
        },
      });
      index.load(
        props.graph.data.nodes.map((node) => ({
          geometry: { coordinates: node.attributes.coordinates },
          properties: node,
        }))
      );
      this.setState({ index });
    }

    if (rebuildIndex || zoom !== this.state.zoom) {
      const nodes = this.state.index.getClusters([-180, -85, 180, 85], zoom);

      // Create node index for faster coordinate search
      const node_index = {};
      for (const node of nodes) {
        const coordinates = node.geometry.coordinates;
        if (node.properties.hasOwnProperty('attributes')) {
          node_index[node.properties.id] = {
            id: node.properties.id,
            coordinates: coordinates,
          };
        } else {
          const ids = node.properties.ids;
          for (const id of ids) {
            if (!node_index[id]) {
              node_index[id] = {
                id: node.id,
                coordinates: coordinates,
              };
            }
          }
        }
      }

      // Add updated coordinates to edges
      const edges = {};
      props.graph.data.edges.forEach((edge) => {
        const source = node_index[edge.from];
        const target = node_index[edge.to];

        if (source && target) {
          if (
            source.coordinates[0] === target.coordinates[0] &&
            source.coordinates[1] === target.coordinates[1]
          ) {
            return;
          } else {
            const key = `${source.id}_${target.id}`;
            if (!edges[key]) {
              edges[key] = {
                ...edge,
                from_coordinates: source.coordinates,
                to_coordinates: target.coordinates,
                count: 1,
              };
            } else edges[key].count++;
          }
        }
      }, []);

      // console.log(
      //   `Number of edge crossings at zoom-level ${zoom} is ${EdgeCrossings(
      //     Object.values(edges)
      //   )}`
      // );

      this.setState({
        nodes,
        edges: Object.values(edges),
        zoom,
      });
    }
  }

  renderLayers() {
    return [
      new LineLayer({
        id: 'edges',
        data: this.state.edges,
        pickable: true,
        extensions: [new BrushingExtension()],
        brushingEnabled: this.props.brushing,
        brushingRadius: 500000,
        brushingTarget: 'source_target',
        getWidth: (d) => Math.sqrt(d.count) * 0.5,
        getSourcePosition: (d) => d.from_coordinates,
        getTargetPosition: (d) => d.to_coordinates,
        getColor: (d) =>
          getColor(d.attributes.prediction, this.props.colorScale),
      }),
      new ScatterplotLayer({
        id: 'nodes',
        data: this.state.nodes,
        opacity: 1,
        pickable: true,
        interactive: true,
        filled: true,
        radiusScale: 3,
        radiusMinPixels: 5,
        radiusMaxPixels: 20,
        getRadius: (d) =>
          d.hasOwnProperty('type') ? d.properties.ids.length * 1000 : 10,
        getFillColor: (d) => (d.properties.cluster ? [250, 0, 0] : [0, 0, 255]),
        getPosition: (d) => d.geometry.coordinates,
      }),
    ];
  }
}
